#ifndef MONTJOIE_FILE_FINITE_ELEMENT_HCURL_AXI_INLINE_CXX

namespace Montjoie
{
  
  inline ElementReference<Dimension2, 4>::~ElementReference<Dimension2, 4>() 
  {
  }
  
  
  inline int ElementReference<Dimension2, 4>::GetNbDof() const
  {
    return nb_dof_loc; 
  }
  
  
  inline int ElementReference<Dimension2, 4>::GetNbDofBoundaries() const
  {
    return nb_dof_boundaries; 
  }
  
  
  inline int ElementReference<Dimension2, 4>::GetNbPointsQuadratureInside() const
  {
    return nb_points_quadrature_inside; 
  }
  
  
    //! returns the number of dofs for Erz
  inline int ElementReference<Dimension2, 4>::GetNbVectorialDof() const
  {
    return fe_hcurl.GetNbDof();
  }
  
  
  //! returns the number of dofs for Etheta
  inline int ElementReference<Dimension2, 4>::GetNbScalarDof() const
  {
    return fe_h1.GetNbDof();
  }
  
  
  //! returns the number of dofs for H
  inline int ElementReference<Dimension2, 4>::GetNbDofH() const
  {
    return 2*fe_h1.GetNbDof();
  }
  
  
  //! returns the number of dofs for Htheta
  inline int ElementReference<Dimension2, 4>::GetNbDofHteta() const
  {
    return fe_hcurl.GetNbDofH();
  }

  
  //! returns the numbers of dofs on boundary num_loc
  inline int ElementReference<Dimension2, 4>::GetNbDofBoundary(int n) const
  {
    return this->EdgesDof.GetM();
  }
  

  //! returns number of the quadrature point k of the edge j
  /*!
    \param[in] k local quadrature point number on the edge
    \param[in] num_loc edge number
   */
  inline int ElementReference<Dimension2, 4>::GetQuadNumber(int num_loc, int k) const
  {
    return num_quad_points_surf(num_loc)(k);
  }

  
  //! returns the dof number of dof j on edge num_loc
  inline int ElementReference<Dimension2, 4>::GetLocalNumber(int num_loc, int j) const
  {
    return this->EdgesDof(j, num_loc);
  }

  
  //! return the dof number of scalar dof j of edge num_loc
  inline int ElementReference<Dimension2, 4>::GetLocalNumberScal(int num_loc, int j) const
  {
    return this->EdgesDof(fe_hcurl.GetOrder() + j, num_loc);
  }
  
  
  //! return the dof number of vectorial dof j of edge num_loc
  inline int ElementReference<Dimension2, 4>::GetLocalNumberVec(int num_loc, int j) const
  {
    return this->EdgesDof(j, num_loc);
  }
  

  //! returns the number of vectorial dofs along an edge
  inline int ElementReference<Dimension2, 4>::GetNbVectorialDofOnEdge() const
  {
    return fe_hcurl.GetNbDofBoundary(0);
  }
  

  //! return true if the dof j is a tangential dof of the edge num_loc
  inline bool ElementReference<Dimension2, 4>::IsTangentialDof(int j, int num_loc) const
  {
    if (j >= PosDofOnFace.GetM())
      return false;
    
    if (PosDofOnFace(j) & power_two_face(num_loc))
      return true;
    
    return false;
  }
  
    
  inline const VectR2& ElementReference<Dimension2, 4>::PointsQuadInsideND() const
  {
    return points2d_inside; 
  }
  
  
  inline int ElementReference<Dimension2, 4>::GetNbPointsDofInside() const
  {
    return nb_points_dof_inside; 
  }
  
  
  inline int ElementReference<Dimension2, 4>::GetNbPointsDofSurface(int num_loc) const
  {
    return num_dof_points_surf(num_loc).GetM(); 
  }
  
  
  inline int ElementReference<Dimension2, 4>::GetPointDofNumber(int num_loc, int k) const
  {
    return num_dof_points_surf(num_loc)(k); 
  }


  inline bool ElementReference<Dimension2, 4>::UseQuadraturePointsForSh() const
  {
    return false; 
  }
  
  
  inline bool ElementReference<Dimension2, 4>::LinearSparseMassMatrix() const
  {
    return false; 
  }
  
  
  inline bool ElementReference<Dimension2, 4>::DiagonalMassMatrix() const
  {
    return false; 
  }
  
  
  inline bool ElementReference<Dimension2, 4>::LumpedMassMatrix() const
  {
    return false; 
  }
  
  
  inline bool ElementReference<Dimension2, 4>::DiscontinuousElement() const
  {
    return false; 
  }


  inline void ElementReference<Dimension2, 4>::SetPiolaTransform(bool flag)
  {
  }
  
  
  inline int ElementReference<Dimension2, 4>::GetTypeIntegrationEdge() const
  {
    return fe_h1.GetTypeIntegrationEdge(); 
  }
  
  
  inline int ElementReference<Dimension2, 4>::GetOrder() const
  {
    return fe_h1.GetOrder(); 
  }

  
  inline void ElementReference<Dimension2, 4>
  ::FjSurfaceElem(const VectR_N& s, SetPoints<Dimension2>& res,
		  const Mesh<Dimension2>& mesh, int nquad, int ne) const
  { 
    fe_h1.FjSurfaceElem(s, res, mesh, nquad, ne); 
  }
  
  
  inline void ElementReference<Dimension2, 4>
  ::DFjSurfaceElem(const VectR_N& s, const SetPoints<Dimension2>& res,
		   SetMatrices<Dimension2>& mat,
		   const Mesh<Dimension2>& mesh, int nquad, int ne) const
  { 
    fe_h1.DFjSurfaceElem(s, res, mat, mesh, nquad, ne); 
  }
  
  
  inline void ElementReference<Dimension2, 4>
  ::FjSurfaceElemDof(const VectR_N& s, SetPoints<Dimension2>& res,
		     const Mesh<Dimension2>& mesh, int nquad, int ne) const
  {
    fe_h1.FjSurfaceElemDof(s, res, mesh, nquad, ne); 
  }
  
  
  inline void ElementReference<Dimension2, 4>
  ::DFjSurfaceElemDof(const VectR_N& s, const SetPoints<Dimension2>& res,
		      SetMatrices<Dimension2>& mat,
		      const Mesh<Dimension2>& mesh, int nquad, int ne) const
  {
    fe_h1.DFjSurfaceElemDof(s, res, mat, mesh, nquad, ne); 
  }
    
  
  inline void ElementReference<Dimension2, 4>
  ::SetInterpolationElement(Mesh<Dimension2>& mesh, MeshNumbering<Dimension2>&)
  {
  }
    

  //! returns the values of basis functions associated with Htheta on a point
  /*!
    \param[in] point point where basis functions are evaluated
    \param[out] phi values of basis functions associated with Htheta
   */
  inline void ElementReference<Dimension2, 4>
  ::ComputeValuesPhiHtetaRef(const R2& point, VectReal_wp& phi) const
  {
    fe_hcurl.ComputeValuesPhiHRef(point, phi);
  }
  
  
  //! computing values of Hrz on quadrature points
  /*!
    \param[in] U components of Hrz on degrees of freedom
    \param[out] V values of Hrz on quadrature points
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyChTransposeH(const Vector1& U, Vector2& V) const
  {
    Mlt(SeldonTrans, ChH, U, V);
  }

  
  //! computing values of Erz on quadrature points
  /*!
    \param[in] U components of Erz on degrees of freedom
    \param[out] V values of Erz on quadrature points
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyChTransposeE(const Vector1& U, Vector2& V) const
  {
    Mlt(SeldonTrans, ChE, U, V);
  }
  
  
  //! integration against basis functions of Erz
  /*!
    \param[in] U \omega_k f(\xi_k) with (omega_k, \xi_k) quadrature formulaes
    \param[out] V V_i = \int f \varphi_i 
    with varphi_i basis function associated with Erz
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyChE(const Vector1& U, Vector2& V) const
  {
    Mlt(ChE, U, V);
  }
  
  
  //! integration against basis functions of Hrz
  /*!
    \param[in] U \omega_k f(\xi_k) with (omega_k, \xi_k) quadrature formulaes
    \param[out] V V_i = \int f \varphi_i 
    with varphi_i basis function associated with Hrz
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyChH(const Vector1& U, Vector2& V) const
  {
    Mlt(ChH, U, V);
  }
  
  
  //! computing values of Htheta on quadrature points
  /*!
    \param[in] U components of Htheta on degrees of freedom
    \param[out] V values of Htheta on quadrature points
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyChTransposeHteta(const Vector1& U, Vector2& V) const
  {
    Mlt(SeldonTrans, ChHteta, U, V);
  }
  
  
  //! computing values of Etheta on quadrature points
  /*!
    \param[in] U components of Etheta on degrees of freedom
    \param[out] V values of Etheta on quadrature points
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyChTransposeEteta(const Vector1& U, Vector2& V) const
  {
    Mlt(SeldonTrans, ChEteta, U, V);
  }
  
  
  //! integration against basis functions of Etheta
  /*!
    \param[in] U \omega_k f(\xi_k) with (omega_k, \xi_k) quadrature formulaes
    \param[out] V V_i = \int f \varphi_i 
    with varphi_i basis function associated with Etheta
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyChEteta(const Vector1& U, Vector2& V) const
  {
    Mlt(ChEteta, U, V);
  }
  
  
  //! integration against basis functions of Htheta
  /*!
    \param[in] U \omega_k f(\xi_k) with (omega_k, \xi_k) quadrature formulaes
    \param[out] V V_i = \int f \varphi_i 
    with varphi_i basis function associated with Htheta
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyChHteta(const Vector1& U, Vector2& V) const
  {
    Mlt(ChHteta, U, V);
  }
  
  
  //! computing curl of Etheta on quadrature points
  /*!
    \param[in] U components of Etheta on degrees of freedom
    \param[out] V curl of Etheta on quadrature points
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyRhTransposeEteta(const Vector1& U, Vector2& V) const
  {
    Mlt(SeldonTrans, RhEteta, U, V);
  }
  
  
  //! computing curl of Erz on quadrature points
  /*!
    \param[in] U components of Etheta on degrees of freedom
    \param[out] V curl of Erz on quadrature points
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyRhTransposeE(const Vector1& U, Vector2& V) const
  {
    Mlt(SeldonTrans, RhE, U, V);
  }


  //! integration against curl of basis functions of Etheta
  /*!
    \param[in] U \omega_k f(\xi_k) with (omega_k, \xi_k) quadrature formulaes
    \param[out] V V_i = \int f curl(\varphi_i) 
    with varphi_i basis function associated with Etheta
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyRhEteta(const Vector1& U, Vector2& V) const
  {
    Mlt(RhEteta, U, V);
  }
  
  
  //! integration against curl of basis functions of Erz
  /*!
    \param[in] U \omega_k f(\xi_k) with (omega_k, \xi_k) quadrature formulaes
    \param[out] V V_i = \int f curl(\varphi_i) 
    with varphi_i basis function associated with Erz
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyRhE(const Vector1& U, Vector2& V) const
  {
    Mlt(RhE, U, V);
  }
  
  
  //! integration against tangential derivative of basis functions of Etheta
  /*!
    \param[in] U \omega_k f(\xi_k) with (omega_k, \xi_k) quadrature formulaes
    \param[out] V V_i = \int_{edge} f d/ds(\varphi_i) 
    \param[in] num_loc edge number
    with varphi_i basis function associated with Etheta
   */
  template<class Vector1, class Vector2>
  inline void ElementReference<Dimension2, 4>
  ::ApplyRhTangential(const Vector1& U, Vector2& V, int num_loc) const
  {
    Mlt(RhSurf(num_loc), U, V);
  }
  
  

  inline bool ElementReference<Dimension2, 4>::OptimizedComputationMassMatrix() const
  {
    return false; 
  }
  
   
  inline bool ElementReference<Dimension2, 4>::SparseMassMatrix() const
  {
    return false; 
  }
  
  
  template<class T0>
  inline void ElementReference<Dimension2, 4>
  ::ApplyRotationCyclic(int n, FftInterface<Complex_wp>& fft_interface,
			TinyVector<T0, 3>& val_u) 
  {
  }
  
  
  template<class T0>
  inline void ElementReference<Dimension2, 4>
  ::ApplyInverseRotationCyclic(int n, FftInterface<Complex_wp>& fft_interface,
			       TinyVector<T0, 3>& val_u) 
  {
  }
    
  
  template<class T>
  inline void ElementReference<Dimension2, 4>
  ::ApplyInverseRotationCyclic(int n, FftInterface<Complex_wp>& fft_interface,
			       Vector<T>& val_u) 
  {
  }


  template<class TypeEltH1, class TypeEltHcurl>
  inline FiniteElementHcurlAxi<TypeEltH1, TypeEltHcurl>::FiniteElementHcurlAxi()
    : ElementReference<Dimension2, 4>(finite_element_h1, finite_element_hcurl, 
				      finite_element_h1.GetGeometricElement())
  {
  }
  
}

#define MONTJOIE_FILE_FINITE_ELEMENT_HCURL_AXI_INLINE_CXX
#endif
